from dune.ufl import Constant
from ufl import as_tensor, as_vector, dot, grad, inner, zero

from ddfem.boundary import BndFlux_c, BndFlux_v, BndValue


def chModel():
    class CHModel:
        dimRange = [1, 1]

        M = Constant(1, name="M")
        lmbda = Constant(1.0e-02, name="lmbda")  # surface parameter
        dt = Constant(5.0e-06, name="dt")  # time step
        theta = Constant(
            0.5, name="theta"
        )  # theta=1 -> backward Euler, theta=0.5 -> Crank-Nicolson

        u_h_n = None
        Du_h_n = None

        def setup(u_h, DDMModel):
            CHModel.u_h_n = u_h.copy()
            CHModel.Du_h_n = lambda t, x: DDMModel.sigma(
                t, x, CHModel.u_h_n, grad(CHModel.u_h_n)
            )

        def energy(U, DU):
            C, MU = as_vector([U[0]]), as_vector([U[1]])
            DC, DMU = as_tensor([DU[0, :]]), as_tensor([DU[1, :]])

            f = 100 * dot(C, C) * dot(as_vector([1]) - C, as_vector([1]) - C)

            return CHModel.lmbda / 2 * inner(DC, DC) + f

        def F_v(t, x, U, DU):
            C, MU = as_vector([U[0]]), as_vector([U[1]])
            DC, DMU = as_tensor([DU[0, :]]), as_tensor([DU[1, :]])
            c_h_n, mu_h_n = as_vector([CHModel.u_h_n[0]]), as_vector([CHModel.u_h_n[1]])
            Du_h_n = CHModel.Du_h_n(t, x)
            Dc_h_n, Dmu_h_n = as_vector([Du_h_n[0, :]]), as_vector([Du_h_n[1, :]])

            mu_mid = (1.0 - CHModel.theta) * Dmu_h_n + CHModel.theta * DMU
            concentration_F_v = CHModel.M * mu_mid
            potential_F_v = CHModel.lmbda * DC

            return as_tensor([concentration_F_v[0, :], potential_F_v[0, :]])

        def S_i(t, x, U, DU):
            C, MU = as_vector([U[0]]), as_vector([U[1]])
            DC, DMU = as_tensor([DU[0, :]]), as_tensor([DU[1, :]])
            c_h_n, mu_h_n = as_vector([CHModel.u_h_n[0]]), as_vector([CHModel.u_h_n[1]])

            concentration_S_i = -(C - c_h_n) / CHModel.dt
            potential_S_i = (
                MU - (200 + 400 * dot(C, C)) * C + as_vector([600 * dot(C, C)])
            )

            return as_vector([concentration_S_i[0], potential_S_i[0]])

        def valF_v(t, x, MU, DMU, n):
            return zero(2)

        boundary = {
            "full": BndFlux_v(valF_v),
        }

    return CHModel



from dune.ufl import Constant
from ufl import as_vector, conditional, div, dot, grad, outer, sqrt

from ddfem.boundary import BndFlux_c, BndFlux_v, BndValue


def adModel(exact, D, withVelocity, inverted):
    class Model:
        solution = exact
        dimRange = 1
        diffFactor = Constant(D)

        # this should probably be a vector of dimRange and then used by
        # componentwise multiplication with (u-g):
        outFactor_i = diffFactor

        if withVelocity:

            outFactor_e = 1

        def diff(t, x):
            return Model.diffFactor * (1 - 0.5 * dot(x, x))

        def F_v(t, x, U, DU):
            return Model.diff(t, x) * DU

        if withVelocity:

            def b(t, x):
                return Constant([0.9, 0.5]) + 3 * as_vector([x[1], -x[0]])

            def F_c(t, x, U):
                return outer(U, Model.b(t, x))

        if exact:

            def S_i(t, x, U, DU):
                return -div(Model.F_v(t, x, exact(t, x), grad(exact(t, x)))) + (
                    exact(t, x) - U
                )

        if exact and withVelocity:

            def S_e(t, x, U, DU):
                return div(Model.F_c(t, x, exact(t, x)))

        if exact:
            valD = BndValue(lambda t, x: exact(t, x))
            valFv = BndFlux_v(
                lambda t, x, U, DU, n: Model.F_v(t, x, exact(t, x), grad(exact(t, x)))
                * n
            )

            if withVelocity:
                valFc = BndFlux_c(lambda t, x, U, n: Model.F_c(t, x, exact(t, x)) * n)
                valN = [valFc, valFv]
            else:
                valN = valFv

            boundary = {
                "sides": valD,
                "ends": valN,
            }
        else:
            x0 = as_vector([-0.5, -0.5])
            bnd = lambda x: conditional(dot(x - Model.x0, x - Model.x0) < 0.15, 10, 0)
            valD = BndValue(lambda t, x, U: as_vector([Model.bnd(x)]))
            boundary = {"full": valD}

        if inverted:
            for i in range(1, 5):
                boundary[i] = valD

    return Model



from ufl import conditional, dot, max_value, min_value, sqrt, ln, exp


# https://en.wikipedia.org/wiki/Smooth_maximum, https://en.wikipedia.org/wiki/LogSumExp
def smax_value(a, b, s):
    return 1 / s * ln(exp(s * a) + exp(s * b))


def smin_value(a, b, s):
    return smax_value(a, b, -s)


def ufl_length(p):
    return sqrt(dot(p, p) + 1e-10)
    # return max_value(sqrt(dot(p, p)), 1e-10)
    # return sqrt(dot(p, p))


def ufl_sign(p):
    if isinstance(p, (float, int)):
        return 1 if p > 0 else -1

    return conditional(p > 0, 1, -1)


def ufl_clamp(p, minimum, maximum):
    if isinstance(p, (float, int)):
        return min(max(p, minimum), maximum)

    # def ufl_max(p1, p2):
    #     return conditional(p2 < p1, p1, p2)

    # def ufl_min(p1, p2):
    #     return conditional(p1 < p2, p1, p2)

    # return ufl_min(ufl_max(p, minimum), maximum)
    # # using min_value/max_value, seems to break shape Pie?
    return min_value(max_value(p, minimum), maximum)


def ufl_cross(p1, p2):
    return p1[0] * p2[1] - p1[1] * p2[0]



from ufl import grad, zero

from .transformer_base import transformer


@transformer
def DDM1(Model):
    class DDModel(Model):
        def S_e_source(t, x, U, DU):
            return DDModel.phi(x) * Model.S_e(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_e_convection(t, x, U, DU):
            return -DDModel.BT.BndFlux_cExt(t, x, U)

        def S_outside(t, x, U, DU):
            return -(
                DDModel.BT.jumpV(t, x, U) * (1 - DDModel.phi(x)) / (DDModel.epsilon**3)
            )

        def S_i_source(t, x, U, DU):
            return DDModel.phi(x) * Model.S_i(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_i_diffusion(t, x, U, DU):
            if DDModel.BT.BndFlux_vExt is not None:
                diffusion = DDModel.BT.BndFlux_vExt(t, x, U, DU)
            else:
                diffusion = zero(U.ufl_shape)
            return diffusion

        def F_c(t, x, U):
            return DDModel.phi(x) * Model.F_c(t, x, U)

        def F_v(t, x, U, DU):
            return DDModel.phi(x) * Model.F_v(t, x, U, DDModel.sigma(t, x, U, DU))

    return DDModel



from ufl import as_vector, conditional, sqrt

from .helpers import ufl_length, ufl_sign
from .primitive_base import ORIGIN, SDF


class Vesica(SDF):
    def __init__(self, radius, distance, smooth_radius, *args, **kwargs):
        """Generates sign distance function and domain coordinates for a Vesica.
        i.e. Union of two circles.
        Centred at (0,0)

        Args:
            radius (float,): Radius of each circle.
            distance (float): Distance of circle center from y-axis.
            smooth_radius (float): Smoothing of domain,so no sharp corner when circles connection.
        """
        assert distance != 0, "This is a circle, use Ball class"
        assert distance < radius, "No shape exists, circles cancel each other out"
        assert (
            smooth_radius * distance < 0
        ), "For a smooth edge, smooth_radius needs to be opposite sign of distance"

        self.radius = radius
        self.distance = distance
        self.smooth_radius = smooth_radius

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Vesica({self.radius}, {self.distance}, {self.smooth_radius}, {self._repr_core()})"

    def sdf(self, x):
        # Note: Ignores z
        x0_abs = abs(x[0])
        x1_abs = abs(x[1])

        b = sqrt((self.radius + self.smooth_radius) ** 2 - self.distance**2)

        circle_coords = as_vector([x0_abs + self.distance, x[1]])

        return (
            conditional(
                (x1_abs - b) * self.distance > x0_abs * b,
                ufl_length(as_vector([x0_abs, x1_abs - b])) * ufl_sign(self.distance),
                ufl_length(circle_coords) - self.radius - self.smooth_radius,
            )
            + self.smooth_radius
        )



import time

import dune.fem
import dune.grid

from .domain import Domain


class DomainDune(Domain):
    def __init__(self, omega, x, gridView, filterTolerance=0):
        super().__init__(omega)
        self.x = x
        self.gridView = gridView
        self.fullSDF = self.gridFunction(self.omega(self.x), name="full-sdf")
        if filterTolerance>0:
            self.gridView = self._filter(tolerance=filterTolerance)
            self.fullSDF = self.gridFunction(self.omega(self.x), name="full-sdf")

        self._phi = None
        self._bndProj = None
        self._extProj = None
        self._bndProjSDFs = {}


    def gridFunction(self, expr, name):
        start_ = time.time()
        gf = dune.fem.function.gridFunction(expr, name=name, gridView=self.gridView)
        # print(f"{name} setup: {time.time() - start_}")
        return gf

    def phi(self, x):
        if self._phi is None:
            p = super().phi(self.x)
            self._phi = self.gridFunction(p, "phidomain")

        return self._phi

    def boundary_projection(self, x):
        if self._bndProj is None:
            p = super().boundary_projection(self.x)
            self._bndProj = self.gridFunction(p, "bndproj")

        return self._bndProj

    def external_projection(self, x):
        if self._extProj is None:
            p = super().external_projection(self.x)
            self._extProj = self.gridFunction(p, "extproj")

        return self._extProj

    def generate_projSDF(self, sdf):
        projSDF = self._bndProjSDFs.get(sdf.name)
        if projSDF is None:
            projSDF = super().generate_projSDF(sdf)
            gf = self.gridFunction(projSDF(self.x), f"sdfproj{sdf.name}")
            self._bndProjSDFs[sdf.name] = lambda x: gf
            projSDF = self._bndProjSDFs[sdf.name]
        return projSDF

    def adapt(self, level, lowerTol=-0.1, upperTol=0.1):
        for _ in range(level):

            def mark(e):
                v = self.fullSDF(e, self.gridView.dimension * [1.0 / 3.0])
                return (
                    dune.grid.Marker.refine
                    if v > lowerTol and v < upperTol
                    else dune.grid.Marker.keep
                )

            self.gridView.hierarchicalGrid.adapt(mark)
            # print(self.gridView.size(0))
            lowerTol /= 2
            upperTol /= 2

    def _filter(self, tolerance):
        sd = self.fullSDF
        tol = tolerance  # * self.epsilon.value: issue here with epsilon being a UFL expression due to CSG approach
        phiFilter = (
            lambda e: sd(e, self.gridView.dimension * [1.0 / 3.0]) < tol
        )  # bary needs fixing for squares
        return dune.fem.view.filteredGridView(
            self.gridView, phiFilter, domainId=1, useFilteredIndexSet=True
        )



from dune.ufl import Constant
from ufl import (
    Identity,
    as_vector,
    conditional,
    det,
    diff,
    div,
    dot,
    grad,
    ln,
    nabla_div,
    nabla_grad,
    outer,
    replace,
    sym,
    tr,
    variable,
    zero,
)
from ufl.algorithms.ad import expand_derivatives

from ddfem.boundary import BndFlux_v, BndValue


def hyModel():
    # https://jsdokken.com/dolfinx-tutorial/chapter2/hyperelasticity.html
    class Model:
        dimRange = 2
        rho = Constant(1000, name="density")  # kg/m3
        g = Constant(9.8, name="gravity")  # m/s2

        body = Constant([0.0, -rho * g], "body")  # N/m3
        vd = Constant([0, 0], "fixedboundary")
        T = Constant([0, 0], "traction")  # Pa

        outFactor_i = Constant(1, "outFactor")

        # Elasticity parameters
        # https://www.efunda.com/formulae/solid_mechanics/mat_mechanics/calc_elastic_constants.cfm
        E = 1e7  # Young modulus ; Pa
        nu = 0.4  # Poisson ratio ;
        lamb = Constant(E * nu / ((1 + nu) * (1 - 2 * nu)), name="Lame_1")  # Pa
        mu = Constant(E / (2 * (1 + nu)), name="Lame_2")  # i.e. Shear modulus ; Pa

        def P(U, DU):
            I = Identity(Model.dimRange)
            # Stress
            # Hyper-elasticity, (compressible neo-Hookean model)
            F = variable(I + grad(U))  # Deformation gradient
            C = F.T * F  # Right Cauchy-Green tensor
            J = det(F)
            Ic = tr(C)  # First Invariant
            mu = Model.mu
            lamb = Model.lamb
            psi = (mu / 2) * (Ic - 3) - mu * ln(J) + (lamb / 2) * (ln(J)) ** 2

            p = diff(psi, F)
            p = expand_derivatives(p)
            p = replace(p, {F: I + DU})
            # p = replace(p, {F: I + grad(U)})

            # Linear-elasticity
            # p = lamb * tr(sym(DU)) * I + 2 * mu * sym(DU)

            return p

        def F_v(t, x, U, DU):
            return Model.P(U, DU)

        def S_i(t, x, U, DU):
            return Model.body

        valD = lambda t, x, U: Model.vd

        valN = lambda t, x, U, DU, n: Model.T

        boundary = {
            "left": BndValue(valD),
            "other": BndFlux_v(valN),
        }
        # boundary = {
        #     "left": BndValue(valD),
        #     "right": BndFlux_v(valN),
        #     "bulk": BndFlux_v(valN),
        # }

    return Model



import inspect

import ufl
from ufl import replace, zero
from ufl.algorithms.ad import expand_derivatives
from ufl.core.expr import Expr

from .geometry.domain import Domain
from .geometry.primitive_base import SDF


class BoundaryCondition:
    def __init__(self, value):
        self.value = value

    def __call__(self, *args, **kwds):
        return self.value(*args, **kwds)


class BndValue(BoundaryCondition):
    def __init__(self, value):
        if isinstance(value, ufl.core.expr.Expr):
            super().__init__(lambda t, x, u: value)
        else:
            num_args = len(inspect.signature(value).parameters)
            if num_args == 1:
                super().__init__(lambda t, x, u: value(x))
            elif num_args == 2:
                super().__init__(lambda t, x, u: value(t, x))
            elif num_args == 3:
                super().__init__(value)
            # elif num_args == 4:
            #     self.SPLIT = True
            #     super().__init__(lambda t, x, u: value(t, x, u, n, n))
            else:
                raise ValueError(f"Boundary has {num_args} arguments.")


class BndFlux_v(BoundaryCondition):
    pass


class BndFlux_c(BoundaryCondition):
    pass


def classify_boundary(boundary_dict):

    boundary_flux_cs = {}  # Fluxes for the advection term
    boundary_flux_vs = {}  # Fluxes for the diffusion term
    boundary_values = {}  # Boundary values for Dirichlet

    for k, f in boundary_dict.items():

        if isinstance(k, (Expr, str)):
            ids = [k]
        elif callable(k):
            ids = [k]
        else:
            try:
                ids = []
                for kk in k:
                    ids += [kk]
            except TypeError:
                ids = [k]

        if isinstance(f, (tuple, list)):
            assert len(f) == 2, "too many boundary fluxes provided"
            if isinstance(f[0], BndFlux_v) and isinstance(f[1], BndFlux_c):
                boundary_flux_vs.update([(kk, f[0]) for kk in ids])
                boundary_flux_cs.update([(kk, f[1]) for kk in ids])

            elif isinstance(f[0], BndFlux_c) and isinstance(f[1], BndFlux_v):
                boundary_flux_vs.update([(kk, f[1]) for kk in ids])
                boundary_flux_cs.update([(kk, f[0]) for kk in ids])

            else:
                raise ValueError("Need AFlux and DFlux")

        elif isinstance(f, BndFlux_v):
            boundary_flux_vs.update([(kk, f) for kk in ids])

        elif isinstance(f, BndFlux_c):
            boundary_flux_cs.update([(kk, f) for kk in ids])

        elif isinstance(f, BndValue):
            boundary_values.update([(kk, f) for kk in ids])

        else:
            raise NotImplementedError(f"unknown boundary type {k} : {f}")

    return boundary_flux_cs, boundary_flux_vs, boundary_values


def boundary_validation(Model, override_boundary_dict=None):
    if override_boundary_dict is not None:
        boundary_dict = override_boundary_dict
    else:
        boundary_dict = Model.boundary

    hasFlux_c = hasattr(Model, "F_c")
    hasFlux_v = hasattr(Model, "F_v")

    boundary_flux_cs, boundary_flux_vs, boundary_values = classify_boundary(
        boundary_dict
    )

    if hasFlux_c and hasFlux_v:
        assert len(boundary_flux_cs) == len(
            boundary_flux_vs
        ), "two bulk fluxes given, but one boundary fluxes provided"

    if not hasFlux_c:
        assert len(boundary_flux_cs) == 0, "No bulk Advection, but boundary flux given"

    if not hasFlux_v:
        assert len(boundary_flux_vs) == 0, "No bulk diffusion, but boundary flux given"

    assert boundary_values.keys().isdisjoint(boundary_flux_cs)
    assert boundary_values.keys().isdisjoint(boundary_flux_vs)

    return boundary_flux_cs, boundary_flux_vs, boundary_values


class BoundaryTerms:
    def __init__(self, Model, domainDescription):
        self.Model = Model

        if isinstance(domainDescription, Domain):
            self.domain = domainDescription
        else:
            self.domain = Domain(domainDescription)

        condition = lambda k: isinstance(k, str) or isinstance(k, SDF)
        self.diffuse = {k: v for k, v in Model.boundary.items() if condition(k)}
        self.physical = {k: v for k, v in Model.boundary.items() if not condition(k)}

        self.boundary_flux_cs, self.boundary_flux_vs, self.boundary_values = (
            boundary_validation(self.Model, override_boundary_dict=self.diffuse)
        )

        self.bV_weight = []
        self.bF_weight = []

        for model_key in self.boundary_values.keys():
            phi_i_proj = self.domain.bndProjSDFs(model_key)
            self.bV_weight.append(phi_i_proj)

        for model_key in {*self.boundary_flux_cs.keys(), *self.boundary_flux_vs.keys()}:
            phi_i_proj = self.domain.bndProjSDFs(model_key)
            self.bF_weight.append(phi_i_proj)

        if not self.boundary_values:
            self.BndValueExt = None

        if not self.boundary_flux_vs:
            self.BndFlux_vExt = None

    def _dbc_total_weight(self, x):
        weight = 1e-10
        for w_func in self.bV_weight:
            weight += w_func(x)
        return weight
    def _total_weight(self, x):
        weight = 1e-10  # tol
        for w_func in self.bV_weight + self.bF_weight:
            weight += w_func(x)
        return weight

    def _boundary_extend(self, g, x):
        g_tmp = expand_derivatives(g)
        return replace(g_tmp, {x: self.domain.boundary_projection(x)})

    # perhaps switch methods around so that BndValueExt is setup and then
    # jumpD does sum(w)*U - BndValueExt. But then the exception needs to be caught...
    def jumpV(self, t, x, U, U1=None):
        jdv = zero(U.ufl_shape)

        if U1 is None:
            U1 = U

        for g_func, w_func in zip(self.boundary_values.values(), self.bV_weight):
            g_tmp = g_func(t, x, U)  # g_func is a callable from self.boundary_values
            g_ext = self._boundary_extend(g_tmp, x)
            jdv += w_func(x) * (U1 - g_ext)

        return jdv / self._total_weight(x)

    def BndValueExt(self, t, x, U):
        # called if self.BndValueExt was not set to None in __init__
        z = zero(U.ufl_shape)
        return -self.jumpV(t, x, U, z)

    def jumpFv(self, t, x, U, DU, Fv):
        # (sigma.n-gN)*ds(N) = - wN ( sigma.Dphi + gN|Dphi| )
        #   = wN ( (-sigma.Dphi) - gN(t,x,-Dphi/|Dphi|)|Dphi| )
        #   = wN ( sigma.sn - gN(t,x,sn) ) with sn = -Dphi
        jdf = zero(U.ufl_shape)

        fv_scaled = Fv * self.domain.scaled_normal(x)
        for g_func, w_func in zip(self.boundary_flux_vs.values(), self.bF_weight):
            g_tmp = g_func(t, x, U, DU, self.domain.normal(x))
            g_ext = self._boundary_extend(g_tmp, x)
            jdf += w_func(x) * (fv_scaled - g_ext * self.domain.surface_delta(x))

        return jdf / self._total_weight(x)
        # return jdf * (self._dbc_total_weight(x) / self._total_weight(x))

    def BndFlux_vExt(self, t, x, U, DU):
        # called if self.BndFlux_vExt was not set to None in __init__
        z = zero(DU.ufl_shape)
        return -self.jumpFv(t, x, U, DU, z)

    def BndFlux_cExt(self, t, x, U):
        jda = zero(U.ufl_shape)

        for g_func, w_func in zip(self.boundary_flux_cs.values(), self.bF_weight):
            g_tmp = g_func(t, x, U, self.domain.normal(x))
            g_ext = self._boundary_extend(g_tmp, x)
            jda += w_func(x) * -g_ext * self.domain.surface_delta(x)

        return -jda / self._total_weight(x)



from ufl import (
    as_matrix,
    as_vector,
    conditional,
    cos,
    grad,
    max_value,
    min_value,
    pi,
    sin,
    tanh,
)

from .helpers import smax_value, smin_value, ufl_length

ORIGIN = as_vector([0, 0])


class SDF:
    def __init__(self, epsilon=None, name=None, children=None):
        self.epsilon = epsilon
        self.child_sdf = children if children else []

        self.name = name
        if self.name is None:
            self.name = repr(self)

    def _repr_core(self):
        return f"epsilon={self.epsilon}, name={self.name}, children={self.child_sdf}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self._repr_core()})"

    def sdf(self, x):
        raise NotImplementedError

    def __call__(self, x):
        return self.sdf(x)

    def search(self, child_name):
        if self.name == child_name:
            return self

        queue = self.child_sdf.copy()

        while queue:
            current_child = queue.pop(0)

            if current_child.name == child_name:
                return current_child

            for child in current_child.child_sdf:
                queue.append(child)

        return None

    def propagate_epsilon(self, epsilon):
        self.epsilon = epsilon
        for child in self.child_sdf:
            child.propagate_epsilon(epsilon)

    def phi(self, x, epsilon=None):
        if not epsilon:
            epsilon = self.epsilon
            assert self.epsilon, "Must define epsilon"
        return 0.5 * (1 - tanh((3 * self.sdf(x) / epsilon)))

    def chi(self, x):
        return conditional(self.sdf(x) <= 0, 1, 0)

    def projection(self, x):
        return -grad(self.sdf(x)) * self.sdf(x)

    def boundary_projection(self, x):
        return x + self.projection(x)

    def external_projection(self, x):
        # return self.chi(x) * x + self.boundary_projection(x) * (1 - self.chi(x))
        return x + self.projection(x) * (1 - self.chi(x))

    def union(self, other):
        return Union(self, other)

    def subtraction(self, other):
        return Subtraction(self, other)

    def intersection(self, other):
        return Intersection(self, other)

    def xor(self, other):
        return Xor(self, other)

    def scale(self, sc):
        return Scale(self, sc)

    def invert(self):
        return Invert(self)

    def rotate(self, angle, radians=True):
        return Rotate(self, angle, radians)

    def translate(self, vector):
        return Translate(self, vector)

    def round(self, sc):
        return Round(self, sc)

    def extrude(self, length, split_ends=False):
        if split_ends:
            from .plane import Plane

            ext = Extrusion(self, length * 2, name=f"{self.name}_ext")
            ext = Translate(ext, [0, 0, -length / 2], name=f"{self.name}_sides")
            bot = Plane([0, 0, -1], 0, name=f"{self.name}_bot")
            top = Plane([0, 0, 1], -length, name=f"{self.name}_top")
            z_interval = bot & top
            ext = ext & z_interval
        else:
            return Extrusion(self, length, name=f"{self.name}_ext")
        return ext

    def revolve(self, offset=0, axis="x"):
        return Revolution(self, offset, axis)

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __sub__(self, other):
        return self.subtraction(other)

    def __xor__(self, other):
        return self.xor(other)

    def __mul__(self, other):
        return self.scale(other)

    def __rmul__(self, other):
        if isinstance(other, (int, float)):
            return self.scale(other)
        raise TypeError(f"Cannot multiply a SDF with {type(other)}")

    # possibly use a 'Constant' but then multiple changes will influence previous usage?
    smoothing = None

    def max_value(a, b):
        if SDF.smoothing is None:
            return max_value(a, b)
        else:
            return smax_value(a, b, SDF.smoothing)

    def min_value(a, b):
        if SDF.smoothing is None:
            return min_value(a, b)
        else:
            return smin_value(a, b, SDF.smoothing)


class BaseOperator(SDF):
    def __init__(self, children, epsilon=None, *args, **kwargs):

        if not epsilon and all(child.epsilon for child in children):
            epsilon = children[0].epsilon
            for child in children:
                epsilon = max_value(epsilon, child.epsilon)

        super().__init__(children=children, epsilon=epsilon, *args, **kwargs)

    def __getitem__(self, key):
        return self.child_sdf[key]


class Union(BaseOperator):
    """Union of two SDFs (OR) - not perfect(negative)"""

    def __init__(self, sdf1, sdf2, *args, **kwargs):
        super().__init__(children=[sdf1, sdf2], *args, **kwargs)

    def sdf(self, x):
        return SDF.min_value(self.child_sdf[0].sdf(x), self.child_sdf[1].sdf(x))


class Subtraction(BaseOperator):
    """Subtraction of two SDFs (difference) - not perfect"""

    def __init__(self, sdf1, sdf2, *args, **kwargs):
        super().__init__(children=[sdf1, sdf2], *args, **kwargs)

    def sdf(self, x):
        return SDF.max_value(self.child_sdf[0].sdf(x), -self.child_sdf[1].sdf(x))


class Intersection(BaseOperator):
    """Intersection of two SDFs (AND) - not perfect"""

    def __init__(self, sdf1, sdf2, *args, **kwargs):
        super().__init__(children=[sdf1, sdf2], *args, **kwargs)

    def sdf(self, x):
        return SDF.max_value(self.child_sdf[0].sdf(x), self.child_sdf[1].sdf(x))


class Xor(BaseOperator):
    """Xor of two SDFs (AND) - perfect"""

    def __init__(self, sdf1, sdf2, *args, **kwargs):
        super().__init__(children=[sdf1, sdf2], *args, **kwargs)

    def sdf(self, x):
        a_x = self.child_sdf[0].sdf(x)
        b_x = self.child_sdf[1].sdf(x)
        return SDF.max_value(SDF.min_value(a_x, b_x), -SDF.max_value(a_x, b_x))


class Invert(BaseOperator):
    """Inverts SDF"""

    def __init__(self, sdf1, *args, **kwargs):
        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        return -self.child_sdf[0].sdf(x)


class Scale(BaseOperator):
    """Scales SDF"""

    def __init__(self, sdf1, scale, *args, **kwargs):
        if not isinstance(scale, (int, float)):
            raise TypeError(f"Cannot scale a SDF with {type(scale)}")
        elif not scale > 0:
            raise ValueError("Cannot scale a SDF with nonpositive")
        else:
            self.scale = scale

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        return self.child_sdf[0].sdf(x / self.scale) * self.scale

    def __repr__(self):
        return f"Scale({self.scale}, {self._repr_core()})"


class Rotate(BaseOperator):
    """Rotates SDF, counterclockwise of orgin"""

    def __init__(self, sdf1, angle, radians=True, *args, **kwargs):
        if not radians:
            angle *= pi / 180
        self.angle = angle

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        c = cos(self.angle)
        s = sin(self.angle)

        r = as_matrix(((c, -s), (s, c)))
        return self.child_sdf[0].sdf(r.T * x)

    def __repr__(self):
        return f"Rotate({self.angle}, {self._repr_core()})"


class Translate(BaseOperator):
    """Translates SDF"""

    def __init__(self, sdf1, vec, *args, **kwargs):
        if isinstance(vec, (list, tuple)):
            vec = as_vector(vec)
        self.vec = vec

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        return self.child_sdf[0].sdf(x - self.vec)

    def __repr__(self):
        return f"Translate({self.vec}, {self._repr_core()})"


class Round(BaseOperator):
    """Rounds SDF"""

    def __init__(self, sdf1, scale, *args, **kwargs):
        assert scale > 0
        self._scale = scale  # careful not to overwrite SDF.scale here

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        return self.child_sdf[0].sdf(x) - self._scale

    def __repr__(self):
        return f"Round({self._scale}, {self._repr_core()})"


class Extrusion(BaseOperator):
    """Extrude SDF"""

    def __init__(self, sdf1, extrude_length, *args, **kwargs):
        self.extrude_length = extrude_length

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        d = self.child_sdf[0].sdf(as_vector([x[0], x[1]]))
        w = abs(x[2]) - self.extrude_length
        return SDF.min_value(SDF.max_value(d, w), 0) + ufl_length(
            as_vector([SDF.max_value(d, 0), SDF.max_value(w, 0)])
        )

    def __repr__(self):
        return f"Extrusion({self.extrude_length}, {self._repr_core()})"


class Revolution(BaseOperator):
    """Revolve SDF"""

    def __init__(self, sdf1, offset, axis, *args, **kwargs):
        self.offset = offset
        assert axis in ["x", "y"], "Can only revolve around the x or y axis"
        self.axis = axis

        super().__init__(children=[sdf1], *args, **kwargs)

    def sdf(self, x):
        if self.axis == "x":
            q = as_vector([x[0], ufl_length(as_vector([x[1], x[2]])) - self.offset])
        elif self.axis == "y":
            q = as_vector([ufl_length(as_vector([x[0], x[2]])) - self.offset, x[1]])
        return self.child_sdf[0].sdf(q)

    def __repr__(self):
        return f"Revolution({self.offset}, {self._repr_core()})"



import numpy as np

try:
    import pygmsh
except ImportError:
    pygmsh = None

try:
    import dune
except ImportError:
    print(
        """
Example code requires dune to run. To install run
pip install dune-fem
    """
    )

from dune.alugrid import aluConformGrid as leafGridView
from dune.fem import adapt, mark, markNeighbors
from dune.fem.function import gridFunction
from dune.fem.space import lagrange
from dune.fem.view import adaptiveLeafGridView
from dune.grid import cartesianDomain
from dune.ufl import Constant, Space
from ufl import SpatialCoordinate, sqrt

from ddfem import geometry as gm
from ddfem.geometry.domain_dune import DomainDune


def getDomain(
    initialRefine,
    version,
    adaptLevels=0,
    epsFactor=4.5,
    smoothing=None,
    *args,
    **kwargs,
):

    gm.SDF.smoothing = smoothing
    shiftx, shifty = sqrt(2) * 1e-6, sqrt(3) * 1e-6
    domain_range = [[-1.7 + shiftx, -1.3 + shifty], [1.7 + shiftx, 1.3 + shifty]]
    initial_gridsize = [170 * 2**initialRefine, 130 * 2**initialRefine]
    h = sqrt(
        ((domain_range[1][0] - domain_range[0][0]) / initial_gridsize[0]) ** 2
        + ((domain_range[1][1] - domain_range[0][1]) / initial_gridsize[1]) ** 2
    )

    def get_eps(h):
        return Constant(epsFactor * h * 0.5 ** (adaptLevels / 2), "epsilon")

    balls = [
        [1, [0, 0], "Center"],
        [0.5, [0, 0.8], "TopCut"],
        [0.5, [0, -0.8], "BotCut"],
        [0.5, [1, 0], "RightAdd"],
        [0.5, [-1, 0], "LeftAdd"],
        [1.4, [0, 0], "Cutoff"],
    ]

    b = [gm.Ball(c[0], c[1], name=c[2]) for c in balls]
    # omega = (b[0] - b[1] - b[2] | b[3] | b[4]) & b[5]

    bulk = b[0] - b[1] - b[2] | b[3] | b[4]
    bulk.name = "bulk"

    left_cutoff = b[5] | gm.Plane([-1, 0], -0.4)
    left_cutoff.name = "left"
    right_cutoff = b[5] | gm.Plane([1, 0], -0.4)
    right_cutoff.name = "right"

    cutoff = left_cutoff & right_cutoff  # b[5]

    omega = bulk & cutoff
    omega.name = "full"

    h_max = h * 3
    h_min = h / 2
    radius = 5

    x = SpatialCoordinate(Space(2))
    sdf = omega(x)

    def spacing(x, y, epsilon):
        r_min = epsilon.value
        r_max = radius * epsilon.value
        dist = np.abs(sdf((x, y)))
        if dist <= r_min:
            return geom.characteristic_length_min
        elif dist >= r_max:
            return geom.characteristic_length_max
        else:
            # Linear
            m = (geom.characteristic_length_max - geom.characteristic_length_min) / (
                r_max - r_min
            )
            return m * (dist - r_min) + geom.characteristic_length_min

    if version == "cartesian":
        domain = cartesianDomain(*domain_range, initial_gridsize)
        epsilon = get_eps(h)

    elif version == "fitted":
        if pygmsh is None:
            raise AttributeError("'fitted' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h_max
            geom.characteristic_length_min = h_min
            epsilon = get_eps(h_min)

            disks = [geom.add_disk([c[1][0], c[1][1], 0.0], c[0]) for c in balls]

            disk_removed1 = geom.boolean_difference(disks[0], disks[1])
            disk_removed2 = geom.boolean_difference(disk_removed1, disks[2])

            disk_add = geom.boolean_union([disk_removed2, disks[3], disks[4]])
            shape = geom.boolean_intersection([disk_add, disks[5]])

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y, epsilon),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }

    elif version == "dune_adaptive":
        gridsize = [int(j * h / h_max) for j in initial_gridsize]
        domain = cartesianDomain(*domain_range, gridsize)

    elif version == "gmsh_adaptive":
        if pygmsh is None:
            raise AttributeError("'gmsh_adaptive' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h_max
            geom.characteristic_length_min = h_min
            epsilon = get_eps(h_min)

            geom.add_rectangle(
                [domain_range[0][0], domain_range[0][1], 0.0],
                domain_range[1][0] - domain_range[0][0],
                domain_range[1][1] - domain_range[0][1],
            )

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y, epsilon),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }
    else:
        raise ValueError("invalid mesh type")

    gridView = adaptiveLeafGridView(leafGridView(domain))

    if version == "dune_adaptive":
        omega.epsilon = get_eps(h_min)
        omega.epsilon.value *= radius
        epsilon_value = omega.epsilon.value

        marker = mark

        refinements = int(2 * np.log2(h_max / h_min))

        region = gridFunction(
            omega.phi(x) * (1 - omega.phi(x)), gridView=gridView
        )  # interface

        for j in range(1, refinements + 1):

            omega.epsilon.value = epsilon_value * j / refinements
            marker(region, 0.00247262315663, maxLevel=refinements)  # 1 epsilon

            adapt(gridView.hierarchicalGrid)

        h_min = h_max * 0.5 ** (j / 2)
        epsilon = get_eps(h_min)

    omega.propagate_epsilon(epsilon)
    domain = omega

    domain = DomainDune(omega, x, gridView)
    domain.adapt(level=adaptLevels)

    print(
        f"h_max={h_max}, h_min={h_min * 0.5 ** (adaptLevels / 2)}, epsilon={epsilon.value}"
    )

    return gridView, domain



import numpy as np
try:
    import pygmsh
except ImportError:
    pygmsh = None

try:
    import dune
except ImportError:
    print("""
Example code requires dune to run. To install run
pip install dune-fem
    """)

from ddfem import geometry as gm
from ddfem.geometry.domain_dune import DomainDune

from dune.alugrid import aluConformGrid as leafGridView
from dune.fem.view import adaptiveLeafGridView
from dune.grid import cartesianDomain
from dune.ufl import Constant, Space

from ufl import SpatialCoordinate, sqrt

def getDomain(initialRefine, version, inverted, adaptLevels=0, epsFactor=4.5):

    shiftx, shifty = sqrt(2) * 1e-6, sqrt(3) * 1e-6
    domain_range = [[-0.5 + shiftx, -0.5 + shifty], [0.5 + shiftx, 0.5 + shifty]]
    initial_gridsize = [75 * 2**initialRefine] * 2
    h = sqrt(
        ((domain_range[1][0] - domain_range[0][0]) / initial_gridsize[0]) ** 2
        + ((domain_range[1][1] - domain_range[0][1]) / initial_gridsize[1]) ** 2
    )
    epsilon = Constant(epsFactor * h * 0.5 ** (adaptLevels / 2), "epsilon")

    print(f"h={h * 0.5 ** (adaptLevels / 2)}, epsilon={epsilon.value}")

    circles = [
        [0.3, [0.15, 0.15], "b1"],
        [0.3, [-0.15, -0.15], "b2"],
        [0.4, [0, 0], "b3"],
    ]

    b = [gm.Ball(c[0], c[1], epsilon=epsilon, name=c[2]) for c in circles]
    sdfs = [b[0] | b[1], b[2]]
    sdfs[0].name = "sides"
    sdfs[1].name = "ends"
    omega = b[2] # sdfs[0] & sdfs[1]
    if inverted:
        omega = omega.invert()
    omega.name = "full"

    x = SpatialCoordinate(Space(2))
    sdf = omega(x)

    def spacing(x, y):
        r_min = 2*epsilon.value
        r_max = 32 * epsilon.value
        dist = np.abs(sdf((x, y)))
        if dist <= r_min:
            return geom.characteristic_length_min
        elif dist >= r_max:
            return geom.characteristic_length_max
        else:
            # Linear
            m = (geom.characteristic_length_max - geom.characteristic_length_min) / (
                r_max - r_min
            )
            return m * (dist - r_min) + geom.characteristic_length_min

    if version == "cartesian":
        domain = cartesianDomain(*domain_range, initial_gridsize)

    elif version == "fitted":
        if pygmsh is None:
            raise AttributeError("'fitted' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h
            geom.characteristic_length_min = h

            disks = geom.add_disk([circles[2][1][0], circles[2][1][1], 0.0], circles[2][0])
                    # [geom.add_disk([c[1][0], c[1][1], 0.0], c[0]) for c in circles]

            # ds = geom.boolean_union([disks[0], disks[1]])
            shape = disks # geom.boolean_intersection([ds, disks[2]])
            if inverted:
                rectangle = geom.add_rectangle(
                    [domain_range[0][0], domain_range[0][1], 0.0],
                    domain_range[1][0] - domain_range[0][0],
                    domain_range[1][1] - domain_range[0][1],
                )
                geom.boolean_difference(rectangle, shape)

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }

    elif version == "adaptive":
        if pygmsh is None:
            raise AttributeError("'adaptive' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h
            geom.characteristic_length_min = h

            geom.add_rectangle(
                [domain_range[0][0], domain_range[0][1], 0.0],
                domain_range[1][0] - domain_range[0][0],
                domain_range[1][1] - domain_range[0][1],
            )

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }

    gridView = adaptiveLeafGridView(leafGridView(domain))

    domain = DomainDune(omega, x, gridView)
    domain.adapt(level=adaptLevels)
    gridView.plot()

    return gridView, domain



from matplotlib import pyplot as plt
from dune.fem.function import gridFunction
from ufl import SpatialCoordinate, triangle

def plotSolution(domain,uh,figsize, **kwargs):
    x = SpatialCoordinate(triangle)
    fig, axs = plt.subplots(1,2, figsize=[2*figsize[0],figsize[1]])
    uh.plot(figure=(fig,axs[0]),**kwargs)
    gridFunction(uh*domain.chi(x)).plot(figure=(fig,axs[1]),**kwargs)
    for a in axs:
        gridFunction(domain.phi(x), gridView=uh.space.gridView).plot(
             onlyContours=True,contours=[0.5],gridLines=None,
             contourWidth=2, contourColor="black", figure=(fig,a))



from dune.ufl import Constant
from ufl import as_vector, div, exp, grad, inner, sqrt

from ddfem.boundary import BndFlux_v, BndValue


def fhModel(inverted):
    class Model:
        dimRange = 1
        outFactor_i = Constant(1, "outFactor")

        def initial(x):
            return 1 / 2 * (x[0] ** 2 + x[1] ** 2) - 1 / 3 * (x[0] ** 3 - x[1] ** 3) + 1

        def exact(t, x):
            return as_vector([exp(-2 * t) * (Model.initial(x) - 1) + 1])

        def K(U, DU):
            # DU = exp(-2 * t) * (x[0] - x[0]^2, x[1] + x[1]^2)
            return 2 / (1 + sqrt(1 + 4 * sqrt(inner(DU, DU))))

        def F_v(t, x, U, DU):
            return Model.K(U, DU) * DU

        def S_i(t, x, U, DU):
            return -div(
                Model.F_v(t, x, Model.exact(t, x), grad(Model.exact(t, x)))
            ) + as_vector([-2 * exp(-2 * t) * (Model.initial(x) - 1)])

        valD = BndValue(lambda t, x, U: Model.exact(t, x))

        valN = BndFlux_v(
            lambda t, x, U, DU, n: Model.F_v(
                t, x, Model.exact(t, x), grad(Model.exact(t, x))
            )
            * n
        )

        boundary = {
            "sides": valD,
            "ends": valN,
        }

        if inverted:
            for i in range(1, 5):
                boundary[i] = valD

    return Model



from ufl import as_vector, conditional, max_value, min_value

from .helpers import ufl_length
from .primitive_base import ORIGIN, SDF


class Box(SDF):
    def __init__(self, width, height, center=ORIGIN, *args, **kwargs):
        self.width = width
        self.height = height

        if isinstance(center, (list, tuple)):
            center = as_vector(center)
        self.center = center

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Box({self.width}, {self.height}, {self.center}, {self._repr_core()})"

    def sdf(self, x):
        # Note: Ignores z
        # shift x
        center_x = x - self.center
        # aux functions
        w0 = abs(center_x[0]) - self.width / 2
        w1 = abs(center_x[1]) - self.height / 2

        g = max_value(w0, w1)

        q = as_vector([max_value(w0, 0), max_value(w1, 0)])

        return conditional(g > 0, ufl_length(q), g)



from ufl import as_matrix, diff, dot, grad, inner, replace, variable, zero
from ufl.algorithms.ad import expand_derivatives

from ..boundary import BndFlux_c, BndFlux_v, BndValue, BoundaryTerms


def pretransformer(Model, domainDescription):
    class DDBase(Model):
        BT = BoundaryTerms(Model, domainDescription)
        boundary = BT.physical
        domain = BT.domain

        hasFlux_c = hasattr(Model, "F_c")
        hasFlux_v = hasattr(Model, "F_v")
        hasSource_i = hasattr(Model, "S_i")
        hasSource_e = hasattr(Model, "S_e")

        hasOutFactor_i = hasattr(Model, "outFactor_i")
        hasOutFactor_e = hasattr(Model, "outFactor_e")

        if BT.BndValueExt is not None:
            boundary[lambda x: DDBase.domain.chi(x) < 0.5] = BndValue(BT.BndValueExt)

            assert (
                hasOutFactor_i or hasOutFactor_e
            ), "Dirichlet boundary requires the attribute outFactor_i or outFactor_e for outside term scaling"
        else:

            if hasFlux_c:
                valFc = BndFlux_c(lambda t, x, U, n: -DDBase.BT.BndFlux_cExt(t, x, U))
            if hasFlux_v:
                valFv = BndFlux_v(
                    lambda t, x, U, DU, n: DDBase.BT.BndFlux_vExt(t, x, U, DU)
                )
            if hasFlux_c and hasFlux_v:
                valN = [valFc, valFv]
            elif hasFlux_c:
                valN = valFc
            elif hasFlux_v:
                valN = valFv
            boundary[lambda x: DDBase.domain.chi(x) < 0.5] = valN

        phi = domain.phi
        epsilon = domain.omega.epsilon
        ep = domain.external_projection

        def sigma(t, x, U, DU=None):
            if DU:
                return DU
            return grad(U)

        if hasSource_e:

            def S_e(t, x, U, DU):
                return replace(
                    expand_derivatives(Model.S_e(t, x, U, DU)),
                    {x: DDBase.ep(x)},
                )

        if hasSource_i:

            def S_i(t, x, U, DU):
                return replace(
                    expand_derivatives(Model.S_i(t, x, U, DU)),
                    {x: DDBase.ep(x)},
                )

        if hasFlux_c:

            def F_c(t, x, U):
                # return Model.F_c(t, DDBase.ep(x), U)

                return replace(
                    expand_derivatives(Model.F_c(t, x, U)),
                    {x: DDBase.ep(x)},
                )

        if hasFlux_v:

            def F_v(t, x, U, DU):
                # return Model.F_v(t, DDBase.ep(x), U, DU)

                return replace(
                    expand_derivatives(Model.F_v(t, x, U, DU)),
                    {x: DDBase.ep(x)},
                )

        # U_t + div[F_c(x,t,U) - F_v(x,t,U,grad[U]) ] = S(x,t,U, grad[U]).

        def F_c_lin(t, x, U):
            U = variable(U)
            d = diff(Model.F_c(t, x, U), U)
            d = expand_derivatives(d)
            return d

        # U.ufl_shape == (1,)
        # F_c(U).ufl_shape == (1, 2,)
        # diff(F_c(U), U).ufl_shape == (1, 2, 1)
        # n.ufl_shape == (2,)
        #
        # s, t = F_c(U).ufl_shape
        # f_c = as_matrix([[dot(d[i, j, :], U) for j in range(t)] for i in range(s)])
        #
        # w, = U.ufl_shape
        # convec = as_vector([dot([f_c[w, :], n) for i in range(w)]) # f_c * n
        #
        # switch order

        def F_c_lin_mult(t, x, U, n):
            G = DDBase.F_c_lin(t, x, U)
            # try:
            #     d = dot(G, n)
            #     print("F_c dot")
            #     return d
            # except:
            m, d, m_ = G.ufl_shape
            return as_matrix(
                [[dot(G[i, :, k], n) for k in range(m_)] for i in range(m)]
            )

        def F_v_lin(t, x, U, DU):
            DU = variable(DU)
            d = diff(Model.F_v(t, x, U, DU), DU)
            d = expand_derivatives(d)
            return d

        def F_v_lin_mult(t, x, U, DU, v):
            G = DDBase.F_v_lin(t, x, U, DU)
            # try:
            #     d = dot(G, v)
            #     print("F_v dot")
            #     return d
            # except:
            m, d = v.ufl_shape
            return as_matrix(
                [[inner(G[i, k, :, :], v) for k in range(d)] for i in range(m)]
            )

    return DDBase


def posttransformer(DDModel):
    class DDM(DDModel):
        if DDModel.hasSource_e or DDModel.hasFlux_c or DDModel.hasOutFactor_e:

            def S_e(t, x, U, DU):
                total = zero(U.ufl_shape)

                if DDModel.hasOutFactor_e:
                    total += DDModel.outFactor_e * DDModel.S_outside(t, x, U, DU)
                if DDModel.hasSource_e:
                    total += DDModel.S_e_source(t, x, U, DU)
                if DDModel.hasFlux_c:
                    total += DDModel.S_e_convection(t, x, U, DU)
                return total

        else:
            try:
                del DDModel.S_e
            except AttributeError:
                pass

        if DDModel.hasSource_i or DDModel.hasFlux_v or DDModel.hasOutFactor_i:

            def S_i(t, x, U, DU):
                total = zero(U.ufl_shape)

                if DDModel.hasOutFactor_i:
                    total += DDModel.outFactor_i * DDModel.S_outside(t, x, U, DU)
                if DDModel.hasSource_i:
                    total += DDModel.S_i_source(t, x, U, DU)
                if DDModel.hasFlux_v:
                    total += DDModel.S_i_diffusion(t, x, U, DU)
                return total

        else:
            try:
                del DDModel.S_i
            except AttributeError:
                pass

        if DDModel.hasFlux_c:

            def F_c(t, x, U):
                return DDModel.F_c(t, x, U)

        else:
            try:
                del DDModel.F_c
            except AttributeError:
                pass

        if DDModel.hasFlux_v:

            def F_v(t, x, U, DU):
                return DDModel.F_v(t, x, U, DU)

        else:
            try:
                del DDModel.F_v
            except AttributeError:
                pass

    return DDM


def transformer(transformer):
    def _transformer(OriginalModel, domainDescription):
        PreModel = pretransformer(OriginalModel, domainDescription)
        Model = transformer(PreModel)
        return posttransformer(Model)

    return _transformer
